﻿EnableExplicit

CompilerIf Defined(INCLUDE_QDOC, #PB_Constant)=0
#INCLUDE_QDOC=1
;/////////////////////////////////////////////////////////////////////////////////
;***qDoc***  - Version 1.2.
;============
;   Compound documents.
;
;   Created with Purebasic 5.21 LTS for Windows.
;
;   Platforms:  All (tested on Windows and Mac OSX).
;               Unicode compatible.
;               Not safe to access a document through multiple threads without mutex protection.
;
;   Thanks to : Kiffi for the code allowing for the swift backing up of SQLite databases.
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;NOTES :
;
;   i)    This library uses SQLite to implement compound file documents (similar to our comDoc utility) and is thus fully cross-platform.
;         A document can contain multiple files accessed with methods similar to PB's file commands. 
;
;   ii)   There is no 'catch document' facility (because 'static' memory cannot be used for a dynamic database!)
;         Applications wishing for this kind of facility will have to first unload the document to a temporary disc file etc.
;
;   iii)  File names (within documents) are not case-sensitive. E.g. "myFile" and "myfile" denote the same file etc.
;
;   iv)   Each individual file can have multiple 'read-only' instances, but only one instance can have that same file open for read and write operations.
;         File deletions will fail if the file is open.
;
;   v)    All files remaining open are automatically closed and tidied up when the parent document object is destroyed with the \Close() method.
;         You can of course close files yourself at any time.
;
;   vi)   The advice, when creating documents, is to create a memory based document and then, when all write operations are complete, to save to disc.
;         This is much faster.
;
;   vii)  If writing large strings consider using \WriteOLEString() and \ReadOLEString() as these are much faster to process.
;/////////////////////////////////////////////////////////////////////////////////


UseSQLiteDatabase()


;--------------------IMPORTS.

  ImportC ""
    sqlite3_backup_init(pDest, zDestName.p-utf8, pSource, zSourceName.p-utf8) 
    sqlite3_backup_step(sqlite3_backup, nPage)
    sqlite3_backup_finish(sqlite3_backup)
  EndImport


;--------------------CONSTANTS / STRUCTURES / PROTOTYPES.

  #qDoc_VERSION$ = "1.0"
  #qDoc_KEYVERSION$ = "Created by qDoc version"
  #qDoc_MINPAGESIZE = 1024
  #qDoc_KEYPAGESIZE$ = "Page Size"
  #qDoc_STRINGPAGESIZE = 4096   ;Must be even.


  ;The following structure is used when retrieving dynamic strings.
    Structure qDoc_MemRead
      StructureUnion
        a.a
        u.u
      EndStructureUnion
    EndStructure

  ;The following structure is the class template for the main qDoc_DocumentObject.
    Structure _qDoc_DocumentObject
      *vTable
      ;Document fields.
        fileName$
        dBaseID.i
      ;File list fields.
        ;The following map holds a copy of the qDoc_FileList table for convenience and speed of access. Each entry is indexed on the filename
        ;and holds the size of each file within the document. 
        ;MapSize() gives the number of files present.
        ;Note that the qDoc_FileList table is always kept up to date (in the database) to safeguard against corruption (e.g. following a power failure).
          Map fileList.i()
          Array fileListEnumeration.s(0)
        ;The following 1-based array holds pointers to all objects (_qDoc_FileObject) representing open files. ArraySize() holds the number of open file instances.
          Array openFileObjects.i(0)
      ;Paging fields.
        pageSize.i
    EndStructure

  ;The following structure is the class template for the qDoc_FileObject.
    Structure _qDoc_FileObject
      *vTable
      name$
      *document._qDoc_DocumentObject  ;Pointer to parent document.
      *fileSize.INTEGER               ;Pointer to the relevant entry in the fileList() map. From here we get the file size etc.
      ;Buffers. The page buffer is twice the size of the document's page size to allow for a temporary buffer for 'roll-backs'.
        pageBuffer.i
        stringBuffer.i
      ;Access.
        blnCanWrite.i
        blnPageNeedWriting.i
        loc.i
      ;Errors.
        errorFlag.i
      ;User data.
        userData.i
    EndStructure


;--------------------INTERFACES.

  Interface qDoc_DocumentObject
    Close()
    CreateFile.i(name$)             ;Replaces any existing file of the same name. Method fails if a file with the given name is already open.
                                    ;Returns a qDoc_FileObject pointer if successful to represent a file object with read/write access.
    DeleteFile.i(name$)             ;The function fails if the file is open. Returns non-zero if successful.
    GetDatabaseID.i()               ;Returns the underlying SQLite database #.
    GetFileSize.i(name$)            ;Returns -1 if the file is not found, otherwise returns the current size of the file.
    OpenFile.i(name$)               ;Opens an existing file within the document for read/write. It does not create a new file if a file is not found.
                                    ;Fails if the file is already open for writing.
                                    ;Returns a qDoc_FileObject pointer if successful.
    ReadFile.i(name$)               ;Opens an existing file within the document for read only operations. 
                                    ;Returns a qDoc_FileObject pointer if successful.
    SaveDocumentAs.i(fileName$)     ;Any existing file is deleted. 
                                    ;Returns non-zero if successful.
    SaveToMemoryBasedDocument.i()   ;Returns, if successful, a qDoc_DocumentObject pointer representing a memory based document containing
                                    ;an exact copy of the underlying document. No files will be opened though.
    ;Enumeration methods. 
    ;Between the calls to \BeginFileList() and \FinishFileList(), the list of file names returned is fixed; that is it will not
    ;reflect any new files added or files removed. Use \BeginFileList() again for an up to date list.
      BeginFileList.i()             ;Returns the number of files within the document.
      GetFileListEntry.s(index)     ;Returns either the appropriate filename or an empty string.
      FinishFileList()
  EndInterface

  Interface qDoc_FileObject
    Close()
    ;General file related methods.
      Eof.i()                                   ;Returns #True or #False.
      FileSeek(newPosition)
      FlushBuffers()                            ;No return value.
      Loc.i()
      Lof.i()
    ;Read methods.
      ReadAsciiCharacter.a()
      ReadByte.b()
      ReadCharacter.c()
      ReadData.i(*memoryBuffer, lengthToRead)   ;Returns the number of bytes actually read.
      ReadDouble.d()
      ReadFloat.f()
      ReadInteger.i()
      ReadLong.l()
      ReadOLEString.s()                         ;Reads a string previously written with the \WriteOLEString() method.
      ReadQuad.q()
      ReadString.s(format=0)                    ;Format = #PB_Ascii or #PB_Unicode etc. Leave empty for a format matching the Unicode compiler switch. 
      ReadUnicodeCharacter.u()
      ReadWord.w()
    ;Write methods.
      WriteAsciiCharacter(value.a)
      WriteByte(value.b)
      WriteCharacter(value.c)
      WriteData(*memoryBuffer, length)
      WriteDouble(value.d)
      WriteFloat(value.f)
      WriteInteger(value.i)
      WriteLong(value.l)
      WriteOLEString(text.p-unicode)            ;Writes a string in Unicode format which consists of 4 bytes containing the number of bytes in the string followed by the bytes of the string.
                                                ;No null terminator is included.
      WriteQuad(value.q)
      WriteString(text$, format=0)              ;A null terminator is added. If format = 0, then the string is written in Ascii/Unicode
                                                ;depending on the Unicode compiler switch etc.
      WriteUnicodeCharacter(value.u)
      WriteWord(value.w)
    ;Miscellaneous methods.
      ClearErrorFlag()
      GetErrorFlag.i()
      GetUserData.i()
      SetUserData.i(value)                      ;Returns the existing (previous) user data.
  EndInterface


;--------------------DECLARES.
  Declare qDoc_fileclass_Close(*this._qDoc_FileObject)
  Declare qDoc_fileclass_FlushBuffers(*this._qDoc_FileObject)


;--------------------INTERNAL FUNCTIONS.


;The following internal function returns the number of bytes read.
Procedure.i qDoc_INTERNAL_ReadBytes(*file._qDoc_FileObject, inBuffer, numBytesToRead)
  Protected numBytesRead, numPagesOnDisc, firstPage, lastPage, iPage, pageBoundary, bytesToRead, t1
  If *file\loc < *file\fileSize\i
    ;Adjust the number of bytes to read if matters would extend beyond the end of the file. We do not set the error flag in such cases.
    ;This can be detected, if needed, by looking at the return value.
      If *file\loc + numBytesToRead > *file\fileSize\i
        numBytesToRead = *file\fileSize\i - *file\loc
      EndIf
    ;Determine which pages will need reading from. The first page is of course the 'current' page, a copy of which will reside within the page buffer.
    ;The last page might not require any reading.
      firstPage = *file\loc / *file\document\pageSize 
      lastPage = (*file\loc + numBytesToRead) / *file\document\pageSize
    ;Write the current page out if required.
      If firstPage <> lastPage And *file\blnPageNeedWriting
        *file\blnPageNeedWriting = #False
        SetDatabaseBlob(*file\document\dBaseID, 0, *file\pageBuffer, *file\document\pageSize)
        DatabaseUpdate(*file\document\dBaseID, "UPDATE " + *file\name$ + " SET data = ? WHERE page = " + Str(firstPage) + ";")
      EndIf
    ;Read from the current (first) page which is in memory already.
      pageBoundary = firstPage * *file\document\pageSize
      bytesToRead = *file\document\pageSize - *file\loc + pageBoundary
      If bytesToRead > numBytesToRead
        bytesToRead = numBytesToRead
      EndIf
      CopyMemory(*file\pageBuffer + *file\loc - pageBoundary, inBuffer, bytesToRead)
      numBytesRead + bytesToRead
    ;Now the remaining pages.
      If firstPage <> lastPage
        If DatabaseQuery(*file\document\dBaseID, "SELECT data FROM " + *file\name$ + " WHERE page > " + Str(firstPage) + " AND page <= " + Str(lastPage) + ";")
          For iPage = firstPage+1 To lastPage
            If iPage < lastPage
              If NextDatabaseRow(*file\document\dBaseID)
                t1 = DatabaseColumnSize(*file\document\dBaseID, 0)
                If t1 = *file\document\pageSize
                  GetDatabaseBlob(*file\document\dBaseID, 0, inBuffer + numBytesRead, t1)
                Else
                  FillMemory(inBuffer + numBytesRead, *file\document\pageSize, 0)
                EndIf
              Else
                FillMemory(inBuffer + numBytesRead, *file\document\pageSize, 0)
              EndIf
              numBytesRead + *file\document\pageSize
            Else
              If NextDatabaseRow(*file\document\dBaseID)
                t1 = DatabaseColumnSize(*file\document\dBaseID, 0)
                If t1 = *file\document\pageSize
                  GetDatabaseBlob(*file\document\dBaseID, 0, *file\pageBuffer, t1)
                Else
                  FillMemory(*file\pageBuffer, *file\document\pageSize, 0)
                EndIf
              EndIf
              bytesToRead = numBytesToRead - numBytesRead
              If bytesToRead
                CopyMemory(*file\pageBuffer, inBuffer + numBytesRead, bytesToRead)
              EndIf
              numBytesRead + bytesToRead
            EndIf
          Next
          FinishDatabaseQuery(*file\document\dBaseID)
        EndIf
      EndIf
    ;Reset the file pointer.
      *file\loc + numBytesRead
  EndIf
  ProcedureReturn numBytesRead
EndProcedure


;The following internal function is called whenever data needs to be written to a file within a document.
;It will write directly from the supplied buffer if possible, but occasionally will have to augment this with the file's page buffer.
;Any errors and the write is undone (rolled back) and the file's error flag set.
;Returns the number of bytes written.
Procedure.i qDoc_INTERNAL_WriteBytes(*file._qDoc_FileObject, outBuffer, numBytesToWrite)
  Protected numBytesWritten, numPagesOnDisc, firstPage, lastPage, iPage, pageBoundary, numBytesRemaining, t1
  Protected dbError, tranCount, blnPageMemoryAltered
  If *file\blnCanWrite
  ;Determine which pages will need writing. The first page is of course the 'current' page, a copy of which will reside within the page buffer.
  ;The last page might not require any writing.
    numPagesOnDisc = *file\fileSize\i / *file\document\pageSize + 1
    firstPage = *file\loc / *file\document\pageSize
    lastPage = (*file\loc + numBytesToWrite) / *file\document\pageSize
    pageBoundary = firstPage * *file\document\pageSize
    For iPage = firstPage To lastPage
      If iPage = firstPage
        If *file\loc = pageBoundary And iPage <> lastPage
          dbError + DatabaseUpdate(*file\document\dBaseID, "BEGIN TRANSACTION")
          SetDatabaseBlob(*file\document\dBaseID, 0, outBuffer, *file\document\pageSize)
          dbError + DatabaseUpdate(*file\document\dBaseID, "UPDATE " + *file\name$ + " SET data = ? WHERE page = " + Str(firstPage) + ";")
          numBytesWritten + *file\document\pageSize
          tranCount = 2
        ElseIf iPage <> lastPage
          ;Copy the page buffer to a temporary store in case we need to rollback.
            CopyMemory(*file\pageBuffer, *file\pageBuffer + *file\document\pageSize, *file\document\pageSize)
            blnPageMemoryAltered = #True
          dbError + DatabaseUpdate(*file\document\dBaseID, "BEGIN TRANSACTION")
          t1 = *file\document\pageSize - *file\loc + pageBoundary
          CopyMemory(outBuffer, *file\pageBuffer + *file\loc - pageBoundary, t1)
          SetDatabaseBlob(*file\document\dBaseID, 0, *file\pageBuffer, *file\document\pageSize)
          dbError + DatabaseUpdate(*file\document\dBaseID, "UPDATE " + *file\name$ + " SET data = ? WHERE page = " + Str(firstPage) + ";")
          numBytesWritten + t1
          tranCount = 2
        Else ;Last page!
          CopyMemory(outBuffer, *file\pageBuffer + *file\loc - pageBoundary, numBytesToWrite)
          *file\blnPageNeedWriting = #True
          numBytesWritten + numBytesToWrite
        EndIf
      ElseIf iPage < lastPage
        SetDatabaseBlob(*file\document\dBaseID, 0, outBuffer + numBytesWritten, *file\document\pageSize)
        If iPage < numPagesOnDisc
          dbError + DatabaseUpdate(*file\document\dBaseID, "UPDATE " + *file\name$ + " SET data = ? WHERE page = " + Str(iPage) + ";")
        Else
          dbError + DatabaseUpdate(*file\document\dBaseID, "INSERT INTO " + *file\name$ + " VALUES (" + Str(iPage) + ", ?);")
        EndIf
        tranCount + 1
        numBytesWritten + *file\document\pageSize
      Else ;Last page which will not equate to the first page and will not require writing to disc because we will not be dealing with the entire page.
        ;Check for a database error.
        If dbError = tranCount
          DatabaseUpdate(*file\document\dBaseID, "COMMIT")
          *file\blnPageNeedWriting = #False
          If iPage < numPagesOnDisc
            If DatabaseQuery(*file\document\dBaseID, "SELECT data FROM " + *file\name$ + " WHERE page = " + Str(lastPage) + ";")
              If NextDatabaseRow(*file\document\dBaseID)
                t1 = DatabaseColumnSize(*file\document\dBaseID, 0)
                If t1 ;This will be no bigger than the document page size.
                  GetDatabaseBlob(*file\document\dBaseID, 0, *file\pageBuffer, t1)
                Else ;No blob written.
                  FillMemory(*file\pageBuffer, *file\document\pageSize, 0)
                EndIf
              Else
                FillMemory(*file\pageBuffer, *file\document\pageSize, 0)
              EndIf
              FinishDatabaseQuery(*file\document\dBaseID)
            EndIf
          Else
            DatabaseUpdate(*file\document\dBaseID, "INSERT INTO " + *file\name$ + " (page) VALUES (" + Str(lastPage) + ");")
            FillMemory(*file\pageBuffer, *file\document\pageSize, 0)
          EndIf
          numBytesRemaining = numBytesToWrite - numBytesWritten
          If numBytesRemaining
            CopyMemory(outBuffer + numBytesWritten, *file\pageBuffer, numBytesRemaining)
            *file\blnPageNeedWriting = #True
            numBytesWritten + numBytesRemaining
          EndIf
        Else
          DatabaseUpdate(*file\document\dBaseID, "ROLLBACK")
          ;Roll back the first page if needed.
            If blnPageMemoryAltered
              CopyMemory(*file\pageBuffer + *file\document\pageSize, *file\pageBuffer, *file\document\pageSize)
            EndIf
        EndIf
      EndIf
      pageBoundary + *file\document\pageSize
    Next
    ;If there were no errors then reset the file pointer and, if required, the file size.
      If dbError = tranCount
        *file\loc + numBytesWritten
        If *file\loc > *file\fileSize\i
          *file\fileSize\i = *file\loc
        EndIf
      Else
        numBytesWritten = 0
        *file\errorFlag = #True
      EndIf
  EndIf
  ProcedureReturn numBytesWritten
EndProcedure


;--------------------qDocDocumentObject CLASS.
;=============================================

;--------------------CLASS CREATION FUNCTIONS.

;The following function creates a new qDocument. Leave the fileName$ parameter empty to create a memory based document.
;Returns, if successful, a qDoc_DocumentObject pointer.
Procedure.i qDoc_CreateDocument(fileName$="", pageSize = #qDoc_MINPAGESIZE)
  Protected *this._qDoc_DocumentObject, fileID, dBaseID, dError = 1
  ;Prepare the way.
    If fileName$
      fileID = CreateFile(#PB_Any, fileName$)
      If fileID
        CloseFile(fileID)
      Else
        ProcedureReturn 0    
      EndIf
    Else
      fileName$ = ":memory:"
    EndIf
  ;Attempt to open the database.
    dBaseID = OpenDatabase(#PB_Any, fileName$, "", "", #PB_Database_SQLite)
    If dBaseID
      ;Attempt to allocate memory for a qDoc_DocumentObject.
        *this = AllocateMemory(SizeOf(_qDoc_DocumentObject))
        If *this
          If pageSize < #qDoc_MINPAGESIZE
            pageSize = #qDoc_MINPAGESIZE
          EndIf
          With *this
            \vTable = ?VTable_documentClass
            \fileName$ = fileName$
            \dBaseID = dBaseID
            \pageSize = pageSize
          EndWith
          ;Write the various tables.
            dError << DatabaseUpdate(*this\dBaseID, "BEGIN TRANSACTION") 
              dError << DatabaseUpdate(*this\dBaseID, "CREATE TABLE qDoc_Master(key TEXT, value TEXT);")
              dError << DatabaseUpdate(*this\dBaseID, "INSERT INTO qDoc_Master VALUES ('" + #qDoc_KEYVERSION$ + "', '" + #qDoc_VERSION$ + "')")
              dError << DatabaseUpdate(*this\dBaseID, "INSERT INTO qDoc_Master VALUES ('" + #qDoc_KEYPAGESIZE$ + "', " + Str(*this\pageSize) + ")")
              dError << DatabaseUpdate(*this\dBaseID, "CREATE TABLE qDoc_FileList(filename TEXT, filesize INTEGER);")
            dError << DatabaseUpdate(*this\dBaseID, "COMMIT")
            If dError = 64
              InitializeStructure(*this, _qDoc_DocumentObject)
            Else
              FreeMemory(*this)
              *this = 0
              Goto qDoc_CreateDocument_L1
            EndIf
        Else
qDoc_CreateDocument_L1:
          CloseDatabase(dBaseID)
        EndIf
    EndIf
  ;In the case that we created a new disc file and something went wrong, we delete the aforementioned file.
    If *this = 0 And fileName$ <> ":memory:"
      DeleteFile(fileName$)
    EndIf
  ProcedureReturn *this
EndProcedure


;The following function opens an existing qDocument. 
;Returns, if successful, a qDoc_DocumentObject pointer.
Procedure.i qDoc_OpenDocument(fileName$)
  Protected *this._qDoc_DocumentObject, dBaseID, pageSize
  ;Attempt to open the database.
    dBaseID = OpenDatabase(#PB_Any, fileName$, "", "", #PB_Database_SQLite)
    If dBaseID
      ;Check for the presence of the master table.
        If DatabaseQuery(dBaseID, "select * from qDoc_Master") 
          ;Retrieve the page size.
            While NextDatabaseRow(dBaseID)
              If GetDatabaseString(dBaseID, 0) = #qDoc_KEYPAGESIZE$
                pageSize = GetDatabaseLong(dBaseID, 1)
                Break
              EndIf             
            Wend
          FinishDatabaseQuery(dBaseID)
          If pageSize >= #qDoc_MINPAGESIZE  ;We have a valid qDocument.
            ;Attempt to allocate memory for a qDoc_DocumentObject.
              *this = AllocateMemory(SizeOf(_qDoc_DocumentObject))
              If *this
                With *this
                  \vTable = ?VTable_documentClass
                  \fileName$ = fileName$
                  \dBaseID = dBaseID
                  \pageSize = pageSize
                EndWith
                ;Populate the *this\fileList() map with a copy of the qDoc_FileList table.
                  If DatabaseQuery(dBaseID, "select * from qDoc_FileList") 
                    InitializeStructure(*this, _qDoc_DocumentObject)
                    While NextDatabaseRow(dBaseID)
                      AddMapElement(*this\fileList(), GetDatabaseString(dBaseID, 0))
                      *this\fileList() = GetDatabaseLong(dBaseID, 1)
                    Wend
                    FinishDatabaseQuery(dBaseID)
                  Else
                    FreeMemory(*this)
                    *this = 0
                    Goto qDoc_OpenDocument_L1
                  EndIf
              Else
                Goto qDoc_OpenDocument_L1
              EndIf
          Else
qDoc_OpenDocument_L1:
            CloseDatabase(dBaseID)
          EndIf
        Else
          CloseDatabase(dBaseID)
        EndIf
    EndIf
  ProcedureReturn *this
EndProcedure


;--------------------CLASS METHODS.

;All open files are automatically closed.
Procedure qDoc_documentclass_Close(*this._qDoc_DocumentObject)
  Protected numFiles, i
  ;Close all open files.
    numFiles = ArraySize(*this\openFileObjects())
    For i = numFiles To 1 Step -1
      qDoc_fileclass_Close(*this\openFileObjects(i))
    Next
  ;Update the qDoc_FileList table.
    DatabaseUpdate(*this\dBaseID, "BEGIN TRANSACTION") 
    ForEach *this\fileList()
      DatabaseUpdate(*this\dBaseID, "UPDATE qDoc_FileList SET filesize = " + Str(*this\fileList()) + " WHERE filename = '" + MapKey(*this\fileList()) + "';")
    Next
    DatabaseUpdate(*this\dBaseID, "COMMIT") 
  CloseDatabase(*this\dBaseID)
  ClearStructure(*this, _qDoc_DocumentObject)
  FreeMemory(*this)
EndProcedure


;The following function creates a new file within the underlying qDocument. 
;Returns, if successful, a qDoc_FileObject pointer.
Procedure.i qDoc_documentclass_CreateFile(*this._qDoc_DocumentObject, fileName$)
  Protected *file._qDoc_FileObject, numFiles, i, dError = 1, *fileList.INTEGER
  If fileName$
    fileName$ = LCase(fileName$)
    ;Check to see if a file with the given name is already open.
      numFiles = ArraySize(*this\openFileObjects())
      For i = 1 To numFiles
        *file = *this\openFileObjects(i)
        If *file\name$ = fileName$
          ProcedureReturn 0 ;A file with the same name is already open.
        EndIf
      Next
    ;Attempt to allocate memory for a qDoc_FileObject.
      *file = AllocateMemory(SizeOf(_qDoc_FileObject))
    If *file
      *file\vTable = ?VTable_fileClass
      *file\document = *this
      *file\blnCanWrite = #True
      *file\blnPageNeedWriting = #False
      *file\pageBuffer = AllocateMemory(*file\document\pageSize<<1)  ;Page buffer.
      If *file\pageBuffer
        *file\stringBuffer = AllocateMemory(#qDoc_STRINGPAGESIZE)
        If *file\stringBuffer
          *file\name$ = fileName$
          ;If a file with the same name exists we empty it.
            dError << DatabaseUpdate(*this\dBaseID, "BEGIN TRANSACTION")
            *fileList = FindMapElement(*this\fileList(), *file\name$)
            If *fileList
              dError << DatabaseUpdate(*this\dBaseID, "DELETE FROM " + *file\name$ + " ;")
              dError << DatabaseUpdate(*this\dBaseID, "UPDATE qDoc_FileList SET filesize = 0 WHERE filename = '" + *file\name$ + "';")
            Else
              ;Create the new table.
                dError << DatabaseUpdate(*this\dBaseID, "CREATE TABLE " + *file\name$ + "(page INTEGER, data BLOB);")
              ;Add a suitable entry to the qDoc_FileList table.
                dError << DatabaseUpdate(*this\dBaseID, "INSERT INTO qDoc_FileList VALUES ('" + *file\name$ + "', 0);")
            EndIf
          ;Write an empty first page to the file. We do this so that all \Write() methods use 'Update' SQL statements for convenience.
            dError << DatabaseUpdate(*this\dBaseID, "INSERT INTO " + *file\name$ + " (page) VALUES (0);")
            If dError = 16
              DatabaseUpdate(*this\dBaseID, "COMMIT")
              ;Adjust the file list map. We add a new item if need be.
                If *fileList
                  *fileList\i = 0
                  *file\fileSize = *fileList
                Else
                  *file\fileSize = AddMapElement(*this\fileList(), *file\name$)
                EndIf
              ;Add a new entry to the openFileObjects() array.
                numFiles + 1
                ReDim *this\openFileObjects(numFiles)
                *this\openFileObjects(numFiles) = *file
            Else
              DatabaseUpdate(*this\dBaseID, "ROLLBACK")
              FreeMemory(*file\stringBuffer)
              Goto qDoc_documentclass_CreateFile_L2
            EndIf
        Else
qDoc_documentclass_CreateFile_L2:
          FreeMemory(*file\pageBuffer)
          Goto qDoc_documentclass_CreateFile_L1
        EndIf
      Else
qDoc_documentclass_CreateFile_L1:
        FreeMemory(*file)
        *file = 0
      EndIf
    EndIf
  EndIf
  ProcedureReturn *file
EndProcedure


;The following function deletes an existing file within the underlying qDocument. The function fails if the file is open.
;Returns non-zero if successful.
Procedure.i qDoc_documentclass_DeleteFile(*this._qDoc_DocumentObject, fileName$)
  Protected result, *file._qDoc_FileObject, numFiles, i, dError = 1
  If fileName$
    fileName$ = LCase(fileName$)
    ;Check to make sure that a file with the given name is not already open.
      numFiles = ArraySize(*this\openFileObjects())
      For i = 1 To numFiles
        *file = *this\openFileObjects(i)
        If *file\name$ = fileName$
          ProcedureReturn 0 ;A file with the same name is already open.
        EndIf
      Next
    ;Proceed only is such a named file exists within the document.
    If FindMapElement(*this\fileList(), fileName$)
      ;We need to drop the table and remove the relevant entry from the qDoc_FileList table
        dError << DatabaseUpdate(*this\dBaseID, "BEGIN TRANSACTION")
        dError << DatabaseUpdate(*this\dBaseID, "DROP TABLE " + fileName$ + ";")
        dError << DatabaseUpdate(*this\dBaseID, "DELETE FROM qDoc_FileList WHERE filename = '" + fileName$ + "';")
        If dError = 8
          DatabaseUpdate(*this\dBaseID, "COMMIT")
          ;Remove the relevant entry from the file list map.
            DeleteMapElement(*this\fileList(), fileName$)
            result = #True
        Else
          DatabaseUpdate(*this\dBaseID, "ROLLBACK")
        EndIf
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure


Procedure.i qDoc_documentclass_GetDatabaseID(*this._qDoc_DocumentObject)
  ProcedureReturn *this\dBaseID
EndProcedure


;Returns -1 if the file is not found.
Procedure.i qDoc_documentclass_GetFileSize(*this._qDoc_DocumentObject, fileName$)
  Protected result = -1
  If FindMapElement(*this\fileList(), LCase(fileName$))
    result = *this\fileList()
  EndIf
  ProcedureReturn result
EndProcedure


;The following function opens an existing file within the underlying qDocument for read/write operations.
;Returns, if successful, a qDoc_FileObject pointer.
Procedure.i qDoc_documentclass_OpenFile(*this._qDoc_DocumentObject, fileName$)
  Protected *file._qDoc_FileObject, numFiles, i, t1, *fileList.INTEGER
  If fileName$
    fileName$ = LCase(fileName$)
    ;Check to see if a file with the same name even exists.
      *fileList = FindMapElement(*this\fileList(), fileName$)
      If *fileList = 0
        ProcedureReturn 0 ;File not found.
      EndIf
    ;Check to see if a file with the given name is already open for write operations
      numFiles = ArraySize(*this\openFileObjects())
      For i = 1 To numFiles
        *file = *this\openFileObjects(i)
        If *file\name$ = fileName$ And *file\blnCanWrite
          ProcedureReturn 0 ;A file with the same name is already open for writing.
        EndIf
      Next
    ;Attempt to allocate memory for a qDoc_FileObject.
      *file = AllocateMemory(SizeOf(_qDoc_FileObject))
    If *file
      *file\vTable = ?VTable_fileClass
      *file\document = *this
      *file\blnCanWrite = #True
      *file\blnPageNeedWriting = #False
      *file\fileSize = *fileList
      *file\pageBuffer = AllocateMemory(*file\document\pageSize<<1)  ;Page buffer.
      If *file\pageBuffer
        *file\stringBuffer = AllocateMemory(#qDoc_STRINGPAGESIZE)
        If *file\stringBuffer
          *file\name$ = fileName$
          ;Read the first page into our memory buffer. The underlying table will have at least one record.
            If DatabaseQuery(*this\dBaseID, "SELECT data FROM " + *file\name$ + " WHERE page = 0")
              If NextDatabaseRow(*this\dBaseID)
                t1 = DatabaseColumnSize(*this\dBaseID, 0)
                If t1 ;This will be no bigger than the document page size.
                  GetDatabaseBlob(*this\dBaseID, 0, *file\pageBuffer, t1)
                EndIf
              Else
                FinishDatabaseQuery(*this\dBaseID)
                Goto qDoc_documentclass_OpenFile_L3
              EndIf
              FinishDatabaseQuery(*this\dBaseID)
              ;Successfully read the first page, now add a new entry to the openFileObjects() array.
                numFiles + 1
                ReDim *this\openFileObjects(numFiles)
                *this\openFileObjects(numFiles) = *file
            Else
qDoc_documentclass_OpenFile_L3:
              FreeMemory(*file\stringBuffer)
              Goto qDoc_documentclass_OpenFile_L2         
            EndIf
        Else
qDoc_documentclass_OpenFile_L2:
          FreeMemory(*file\pageBuffer)
          Goto qDoc_documentclass_OpenFile_L1
        EndIf
      Else
qDoc_documentclass_OpenFile_L1:
        FreeMemory(*file)
        *file = 0
      EndIf
    EndIf
  EndIf
  ProcedureReturn *file
EndProcedure


;The following function opens an existing file within the underlying qDocument for read only operations.
;Returns, if successful, a qDoc_FileObject pointer.
Procedure.i qDoc_documentclass_ReadFile(*this._qDoc_DocumentObject, fileName$)
  Protected *file._qDoc_FileObject, numFiles, i, t1, *fileList.INTEGER
  If fileName$
    fileName$ = LCase(fileName$)
    ;Check to see if a file with the same name even exists.
      *fileList = FindMapElement(*this\fileList(), fileName$)
      If *fileList = 0
        ProcedureReturn 0 ;File not found.
      EndIf
    ;Attempt to allocate memory for a qDoc_FileObject.
      *file = AllocateMemory(SizeOf(_qDoc_FileObject))
    If *file
      *file\vTable = ?VTable_fileClass
      *file\document = *this
      *file\blnCanWrite = #False
      *file\blnPageNeedWriting = #False
      *file\fileSize = *fileList
      *file\pageBuffer = AllocateMemory(*file\document\pageSize<<1)  ;Page buffer.
      If *file\pageBuffer
        *file\stringBuffer = AllocateMemory(#qDoc_STRINGPAGESIZE)
        If *file\stringBuffer
          *file\name$ = fileName$
          ;Read the first page into our memory buffer. The underlying table will have at least one record.
            If DatabaseQuery(*this\dBaseID, "SELECT data FROM " + *file\name$ + " WHERE page = 0")
              If NextDatabaseRow(*this\dBaseID)
                t1 = DatabaseColumnSize(*this\dBaseID, 0)
                If t1 ;This will be no bigger than the document page size.
                  GetDatabaseBlob(*this\dBaseID, 0, *file\pageBuffer, t1)
                EndIf
              Else
                FinishDatabaseQuery(*this\dBaseID)
                Goto qDoc_documentclass_ReadFile_L3
              EndIf
              FinishDatabaseQuery(*this\dBaseID)
              ;Successfully read the first page, now add a new entry to the openFileObjects() array.
                numFiles = ArraySize(*this\openFileObjects()) + 1
                ReDim *this\openFileObjects(numFiles)
                *this\openFileObjects(numFiles) = *file
            Else
qDoc_documentclass_ReadFile_L3:
              FreeMemory(*file\stringBuffer)
              Goto qDoc_documentclass_ReadFile_L2
            EndIf
        Else
qDoc_documentclass_ReadFile_L2:
          FreeMemory(*file\pageBuffer)
          Goto qDoc_documentclass_ReadFile_L1
        EndIf
      Else
qDoc_documentclass_ReadFile_L1:
        FreeMemory(*file)
        *file = 0
      EndIf
    EndIf
  EndIf
  ProcedureReturn *file
EndProcedure


Procedure.i qDoc_documentclass_SaveDocumentAs(*this._qDoc_DocumentObject, fileName$)
  Protected result, fileID, dBaseID, backUp, numFiles, i
  fileID = CreateFile(#PB_Any, fileName$)
  If fileID
    CloseFile(fileID)
  Else
    ProcedureReturn 0    
  EndIf
  ;Attempt to open the database.
    dBaseID = OpenDatabase(#PB_Any, fileName$, "", "", #PB_Database_SQLite)
  If dBaseID
    ;Flush all buffers from open files.
      numFiles = ArraySize(*this\openFileObjects())
      For i = numFiles To 1 Step -1
        qDoc_fileclass_FlushBuffers(*this\openFileObjects(i))
      Next
    ;Update the qDoc_FileList table in the original document prior to copying.
      DatabaseUpdate(*this\dBaseID, "BEGIN TRANSACTION") 
      ForEach *this\fileList()
        DatabaseUpdate(*this\dBaseID, "UPDATE qDoc_FileList SET filesize = " + Str(*this\fileList()) + " WHERE filename = '" + MapKey(*this\fileList()) + "';")
      Next
      DatabaseUpdate(*this\dBaseID, "COMMIT") 
    backUp = sqlite3_backup_init(DatabaseID(dBaseID), "main", DatabaseID(*this\dBaseID), "main") 
    If backUp
      sqlite3_backup_step(backUp, -1)
      If sqlite3_backup_finish(backUp) = 0 ;#SQLITE_OK
        result = #True    
      EndIf
    EndIf
    CloseDatabase(dBaseID)
  EndIf
  ProcedureReturn result
EndProcedure


Procedure.i qDoc_documentclass_SaveToMemoryBasedDocument(*this._qDoc_DocumentObject)
  Protected *memory._qDoc_DocumentObject, backUp, numFiles, i
  *memory = qDoc_CreateDocument("", *this\pageSize)
  If *memory
    ;Flush all buffers from open files.
      numFiles = ArraySize(*this\openFileObjects())
      For i = numFiles To 1 Step -1
        qDoc_fileclass_FlushBuffers(*this\openFileObjects(i))
      Next
    backUp = sqlite3_backup_init(DatabaseID(*memory\dBaseID), "main", DatabaseID(*this\dBaseID), "main") 
    If backUp
      sqlite3_backup_step(backUp, -1)
      If sqlite3_backup_finish(backUp) = 0 ;#SQLITE_OK
        ;Copy the file list map.
          CopyMap(*this\fileList(), *memory\fileList()) 
      Else
        Goto qDoc_documentclass_SaveToMemoryBasedDocument_L1
      EndIf
    Else
qDoc_documentclass_SaveToMemoryBasedDocument_L1:
      qDoc_documentclass_Close(*memory)
      *memory = 0
    EndIf
  EndIf
  ProcedureReturn *memory
EndProcedure


;Returns the number of files within the document.
Procedure.i qDoc_documentclass_BeginFileList(*this._qDoc_DocumentObject)
  Protected numFiles, i
  numFiles = MapSize(*this\fileList())
  ReDim *this\fileListEnumeration(numFiles)
  ForEach *this\fileList()
    i + 1
    *this\fileListEnumeration(i) = MapKey(*this\fileList())  
  Next
  ProcedureReturn numFiles
EndProcedure


Procedure.s qDoc_documentclass_GetFileListEntry(*this._qDoc_DocumentObject, index)
  Protected numFiles
  numFiles = ArraySize(*this\fileListEnumeration())
  If numFiles And index >=0 And index < numFiles
    ProcedureReturn *this\fileListEnumeration(index + 1)
  EndIf
EndProcedure


Procedure qDoc_documentclass_FinishFileList(*this._qDoc_DocumentObject)
  Dim *this\fileListEnumeration(0)
EndProcedure


;--------------------qDocFileObject CLASS.
;-========================================

;--------------------CLASS METHODS.

Procedure qDoc_fileclass_Close(*this._qDoc_FileObject)
  Protected pageIndex, numFiles, i, *file._qDoc_FileObject, j
  ;If the file has write permission then we may have to write the current page out to the database.
    qDoc_fileclass_FlushBuffers(*this)
  ;Remove the file from the open file list.
    numFiles = ArraySize(*this\document\openFileObjects())
    For i = 1 To numFiles
      *file = *this\document\openFileObjects(i)
      If *file = *this
        ;Remove the element.
          numFiles - 1
          For j = i To numFiles        
            *this\document\openFileObjects(j) = *this\document\openFileObjects(j+1)
          Next
        ReDim *this\document\openFileObjects(numFiles)
        Break
      EndIf
    Next
  FreeMemory(*this\stringBuffer)
  FreeMemory(*this\pageBuffer)
  ClearStructure(*this, _qDoc_FileObject)
  FreeMemory(*this)
EndProcedure


Procedure.i qDoc_fileclass_Eof(*this._qDoc_FileObject)
  Protected result
  If *this\loc >= *this\fileSize\i
    result = #True
  EndIf
  ProcedureReturn result
EndProcedure


Procedure qDoc_fileclass_FileSeek(*this._qDoc_FileObject, newPosition)
  Protected currentPage, newPage, t1
  ;Check the value of newPosition.
    If newPosition < 0
      newPosition = 0
    ElseIf newPosition > *this\fileSize\i
      newPosition = *this\fileSize\i
    EndIf
  If newPosition <> *this\loc
    currentPage = *this\loc / *this\document\pageSize
    newPage = newPosition / *this\document\pageSize
    If newPage <> currentPage
      ;We have to load up the new page. First decide if the existing page needs writing.    
        If *this\blnPageNeedWriting
          *this\blnPageNeedWriting = #False
          SetDatabaseBlob(*this\document\dBaseID, 0, *this\pageBuffer, *this\document\pageSize)
          DatabaseUpdate(*this\document\dBaseID, "UPDATE " + *this\name$ + " SET data = ? WHERE page = " + Str(currentPage) + ";")
        EndIf
      ;Now load the new page.
        If DatabaseQuery(*this\document\dBaseID, "SELECT data FROM " + *this\name$ + " WHERE page = " + Str(newPage) + ";")
          If NextDatabaseRow(*this\document\dBaseID)
            t1 = DatabaseColumnSize(*this\document\dBaseID, 0)
            If t1 ;This will be no bigger than the document page size.
              GetDatabaseBlob(*this\document\dBaseID, 0, *this\pageBuffer, t1)
            Else ;No blob written.
              FillMemory(*this\pageBuffer, *this\document\pageSize, 0)
            EndIf
          Else
            FillMemory(*this\pageBuffer, *this\document\pageSize, 0)
          EndIf
          FinishDatabaseQuery(*this\document\dBaseID)
        Else
          FillMemory(*this\pageBuffer, *this\document\pageSize, 0)
        EndIf
    EndIf
    *this\loc = newPosition
  EndIf
EndProcedure


Procedure qDoc_fileclass_FlushBuffers(*this._qDoc_FileObject)
  Protected pageIndex
  If *this\blnPageNeedWriting ;*this\blnCanWrite will automatically be non-zero in this case.
    ;Calculate the page index.
      pageIndex = *this\loc / *this\document\pageSize
    SetDatabaseBlob(*this\document\dBaseID, 0, *this\pageBuffer, *this\document\pageSize)
    DatabaseUpdate(*this\document\dBaseID, "UPDATE " + *this\name$ + " set data = ? where page = " + Str(pageIndex) + ";")
    *this\blnPageNeedWriting = #False
  EndIf
EndProcedure


Procedure.i qDoc_fileclass_Loc(*this._qDoc_FileObject)
  ProcedureReturn *this\loc
EndProcedure


Procedure.i qDoc_fileclass_Lof(*this._qDoc_FileObject)
  ProcedureReturn *this\fileSize\i
EndProcedure


Procedure.a qDoc_fileclass_ReadAsciiCharacter(*this._qDoc_FileObject)
  Protected buffer.a
  qDoc_INTERNAL_ReadBytes(*this, @buffer, SizeOf(ASCII))
  ProcedureReturn buffer
EndProcedure


Procedure.b qDoc_fileclass_ReadByte(*this._qDoc_FileObject)
  Protected buffer.b
  qDoc_INTERNAL_ReadBytes(*this, @buffer, SizeOf(BYTE))
  ProcedureReturn buffer
EndProcedure


Procedure.c qDoc_fileclass_ReadCharacter(*this._qDoc_FileObject)
  Protected buffer.c
  qDoc_INTERNAL_ReadBytes(*this, @buffer, SizeOf(CHARACTER))
  ProcedureReturn buffer
EndProcedure


;Returns the number of bytes actually read.
Procedure.i qDoc_fileclass_ReadData(*this._qDoc_FileObject, *memoryBuffer, lengthToRead)
  Protected numBytesRead
  If *memoryBuffer And lengthToRead
    numBytesRead = qDoc_INTERNAL_ReadBytes(*this, *memoryBuffer, lengthToRead)
  EndIf
  ProcedureReturn numBytesRead
EndProcedure


Procedure.d qDoc_fileclass_ReadDouble(*this._qDoc_FileObject)
  Protected buffer.d
  qDoc_INTERNAL_ReadBytes(*this, @buffer, SizeOf(DOUBLE))
  ProcedureReturn buffer
EndProcedure


Procedure.f qDoc_fileclass_ReadFloat(*this._qDoc_FileObject)
  Protected buffer.f
  qDoc_INTERNAL_ReadBytes(*this, @buffer, SizeOf(FLOAT))
  ProcedureReturn buffer
EndProcedure


Procedure.i qDoc_fileclass_ReadInteger(*this._qDoc_FileObject)
  Protected buffer.i
  qDoc_INTERNAL_ReadBytes(*this, @buffer, SizeOf(INTEGER))
  ProcedureReturn buffer
EndProcedure


Procedure.l qDoc_fileclass_ReadLong(*this._qDoc_FileObject)
  Protected buffer.l
  qDoc_INTERNAL_ReadBytes(*this, @buffer, SizeOf(LONG))
  ProcedureReturn buffer
EndProcedure


Procedure.s qDoc_fileclass_ReadOLEString(*this._qDoc_FileObject)
  Protected text$, numBytes, buffer
  ;First read the 4-bytes preceeding the string bytes as this contains the length (in bytes) of the string.
    If qDoc_INTERNAL_ReadBytes(*this, @numBytes, SizeOf(LONG)) = SizeOf(LONG) And numBytes
      buffer = AllocateMemory(numBytes + SizeOf(UNICODE))
      If buffer
        qDoc_INTERNAL_ReadBytes(*this, buffer, numBytes)
        text$ = PeekS(buffer, -1, #PB_Unicode) 
        FreeMemory(buffer)      
      Else
        *this\errorFlag = #True
      EndIf          
    EndIf
  ProcedureReturn text$
EndProcedure


Procedure.q qDoc_fileclass_ReadQuad(*this._qDoc_FileObject)
  Protected buffer.q
  qDoc_INTERNAL_ReadBytes(*this, @buffer, SizeOf(QUAD))
  ProcedureReturn buffer
EndProcedure


;The following attempts to read a string directly into the file's dynamic string buffer. If we encounter problems allocating additional memory
;then a truncated string is returned.
Procedure.s qDoc_fileclass_ReadString(*this._qDoc_FileObject, format=0)
  Protected text$, charSize, t1, numBytes, byteLen, bufferSize, t2, blnError, dummy
  Protected *memReader.qDoc_MemRead
  If format <> #PB_Ascii And format <> #PB_UTF8 And format <> #PB_Unicode
    CompilerIf #PB_Compiler_Unicode
      format = #PB_Unicode
    CompilerElse
      format = #PB_Ascii
    CompilerEndIf
  EndIf
  If format = #PB_Unicode
    charSize = 2
  Else
    charSize = 1
  EndIf
  ;Fetch bytes either 1 byte or 2 bytes at a time depending on the string format.
    bufferSize = MemorySize(*this\stringBuffer)
    *memReader = *this\stringBuffer
    t1 = *this\loc
    While t1 < *this\fileSize\i
      ;Do we need additional memory?
        If byteLen + charSize > bufferSize
          t2 = ReAllocateMemory(*this\stringBuffer, bufferSize + #qDoc_STRINGPAGESIZE)
          If t2
            *this\stringBuffer = t2
            bufferSize + #qDoc_STRINGPAGESIZE
          Else ;A failure to allocate additional memory. We return a truncated string. Need to continue reading the remainder of the string though.
            blnError = #True
            Break
          EndIf
        EndIf
      numBytes = qDoc_INTERNAL_ReadBytes(*this, *memReader, charSize)
      ;String terminator found?
        If (charSize = 1 And *memReader\a = 0) Or (charSize = 2 And *memReader\u = 0)
          Break
        EndIf
      byteLen + numBytes
      *memReader + numBytes
      t1 + numBytes
    Wend
  If blnError
    ;Need to continue reading the remainder of the string (and discard it). We do it in a separate loop to avoid slowing down the above loop.
      While t1 < *this\fileSize\i
        numBytes = qDoc_INTERNAL_ReadBytes(*this, @dummy, charSize)
        ;String terminator found?
          If dummy = 0
            Break
          EndIf
        t1+numBytes
      Wend
  EndIf
  If byteLen
    text$ = PeekS(*this\stringBuffer, byteLen << (charSize-1), format)
    ;Return the file's string buffer back to the default size if we changed it's size.
    If bufferSize <> #qDoc_STRINGPAGESIZE
      t1 = ReAllocateMemory(*this\stringBuffer, #qDoc_STRINGPAGESIZE)
      If t1
        *this\stringBuffer = t1
      EndIf
    EndIf
  EndIf
  ProcedureReturn text$
EndProcedure


Procedure.u qDoc_fileclass_ReadUnicodeCharacter(*this._qDoc_FileObject)
  Protected buffer.u
  qDoc_INTERNAL_ReadBytes(*this, @buffer, SizeOf(UNICODE))
  ProcedureReturn buffer
EndProcedure


Procedure.w qDoc_fileclass_ReadWord(*this._qDoc_FileObject)
  Protected buffer.w
  qDoc_INTERNAL_ReadBytes(*this, @buffer, SizeOf(WORD))
  ProcedureReturn buffer
EndProcedure


Procedure qDoc_fileclass_WriteAsciiCharacter(*this._qDoc_FileObject, value.a)
  qDoc_INTERNAL_WriteBytes(*this, @value, SizeOf(ASCII)) ;This call will set the error flag if appropriate.
EndProcedure


Procedure qDoc_fileclass_WriteByte(*this._qDoc_FileObject, value.b)
  qDoc_INTERNAL_WriteBytes(*this, @value, SizeOf(BYTE)) ;This call will set the error flag if appropriate.
EndProcedure


Procedure qDoc_fileclass_WriteCharacter(*this._qDoc_FileObject, value.c)
  qDoc_INTERNAL_WriteBytes(*this, @value, SizeOf(CHARACTER)) ;This call will set the error flag if appropriate.
EndProcedure


Procedure qDoc_fileclass_WriteData(*this._qDoc_FileObject, *memoryBuffer, length)
  If *memoryBuffer And length
    qDoc_INTERNAL_WriteBytes(*this, *memoryBuffer, length)  ;This call will set the error flag if appropriate.
  EndIf
EndProcedure


Procedure qDoc_fileclass_WriteDouble(*this._qDoc_FileObject, value.d)
  qDoc_INTERNAL_WriteBytes(*this, @value, SizeOf(DOUBLE)) ;This call will set the error flag if appropriate.
EndProcedure


Procedure qDoc_fileclass_WriteFloat(*this._qDoc_FileObject, value.f)
  qDoc_INTERNAL_WriteBytes(*this, @value, SizeOf(FLOAT)) ;This call will set the error flag if appropriate.
EndProcedure


Procedure qDoc_fileclass_WriteInteger(*this._qDoc_FileObject, value.i)
  qDoc_INTERNAL_WriteBytes(*this, @value, SizeOf(INTEGER)) ;This call will set the error flag if appropriate.
EndProcedure


Procedure qDoc_fileclass_WriteLong(*this._qDoc_FileObject, value.l)
  qDoc_INTERNAL_WriteBytes(*this, @value, SizeOf(LONG)) ;This call will set the error flag if appropriate.
EndProcedure


;textPtr will point to a Unicode string (courtesy of a pseudo-type).
Procedure qDoc_fileclass_WriteOLEString(*this._qDoc_FileObject, textPtr)
  Protected *memReader.qDoc_MemRead, numChars
  ;Need to manually trawl the string to determine it's length.
    *memReader = textPtr
    If *memReader
      While *memReader\u
        numChars + 1
        *memReader + SizeOf(UNICODE)
      Wend
    EndIf    
  ;Write the 4 bytes containing the number of bytes in the unicode string.
    numChars << 1
    If qDoc_INTERNAL_WriteBytes(*this, @numChars, SizeOf(LONG)) = SizeOf(LONG) And numChars
      qDoc_INTERNAL_WriteBytes(*this, textPtr, numChars) ;This call will set the error flag if appropriate.
    EndIf
EndProcedure


Procedure qDoc_fileclass_WriteQuad(*this._qDoc_FileObject, value.q)
  qDoc_INTERNAL_WriteBytes(*this, @value, SizeOf(QUAD)) ;This call will set the error flag if appropriate.
EndProcedure


Procedure qDoc_fileclass_WriteString(*this._qDoc_FileObject, text$, format=0)
  Protected numBytes, byteLength, buffer
  If format <> #PB_Ascii And format <> #PB_UTF8 And format <> #PB_Unicode
    CompilerIf #PB_Compiler_Unicode
      format = #PB_Unicode
    CompilerElse
      format = #PB_Ascii
    CompilerEndIf
  EndIf
  byteLength = StringByteLength(text$+"a", format) ;The 'a' takes the place of the terminating null(s).
  buffer = AllocateMemory(byteLength)
  If buffer
    PokeS(buffer, text$, -1, format)  
    qDoc_INTERNAL_WriteBytes(*this, buffer, byteLength) ;This call will set the error flag if appropriate.
    FreeMemory(buffer)
  Else
    *this\errorFlag = #True
  EndIf
EndProcedure


Procedure qDoc_fileclass_WriteUnicodeCharacter(*this._qDoc_FileObject, value.u)
  qDoc_INTERNAL_WriteBytes(*this, @value, SizeOf(UNICODE)) ;This call will set the error flag if appropriate.
EndProcedure


Procedure qDoc_fileclass_WriteWord(*this._qDoc_FileObject, value.w)
  qDoc_INTERNAL_WriteBytes(*this, @value, SizeOf(WORD)) ;This call will set the error flag if appropriate.
EndProcedure


Procedure qDoc_fileclass_ClearErrorFlag(*this._qDoc_FileObject)
  *this\errorFlag = #False
EndProcedure


Procedure.i qDoc_fileclass_GetErrorFlag(*this._qDoc_FileObject)
  ProcedureReturn *this\errorFlag
EndProcedure


Procedure.i qDoc_fileclass_GetUserData(*this._qDoc_FileObject)
  ProcedureReturn *this\userData
EndProcedure


Procedure.i qDoc_fileclass_SetUserData(*this._qDoc_FileObject, value)
  Protected result
  result = *this\userData
  *this\userData = value
  ProcedureReturn result
EndProcedure


;--------------------vTables.

DataSection

  VTable_documentClass:
    Data.i @qDoc_documentclass_Close()
    Data.i @qDoc_documentclass_CreateFile()
    Data.i @qDoc_documentclass_DeleteFile()
    Data.i @qDoc_documentclass_GetDatabaseID()
    Data.i @qDoc_documentclass_GetFileSize()
    Data.i @qDoc_documentclass_OpenFile()
    Data.i @qDoc_documentclass_ReadFile()
    Data.i @qDoc_documentclass_SaveDocumentAs()
    Data.i @qDoc_documentclass_SaveToMemoryBasedDocument()
    Data.i @qDoc_documentclass_BeginFileList()
    Data.i @qDoc_documentclass_GetFileListEntry()
    Data.i @qDoc_documentclass_FinishFileList()

  VTable_fileClass:
    Data.i @qDoc_fileclass_Close()
    Data.i @qDoc_fileclass_Eof()
    Data.i @qDoc_fileclass_FileSeek()
    Data.i @qDoc_fileclass_FlushBuffers()
    Data.i @qDoc_fileclass_Loc()
    Data.i @qDoc_fileclass_Lof()
    Data.i @qDoc_fileclass_ReadAsciiCharacter()
    Data.i @qDoc_fileclass_ReadByte()
    Data.i @qDoc_fileclass_ReadCharacter()
    Data.i @qDoc_fileclass_ReadData()
    Data.i @qDoc_fileclass_ReadDouble()
    Data.i @qDoc_fileclass_ReadFloat()
    Data.i @qDoc_fileclass_ReadInteger()
    Data.i @qDoc_fileclass_ReadLong()
    Data.i @qDoc_fileclass_ReadOLEString()
    Data.i @qDoc_fileclass_ReadQuad()
    Data.i @qDoc_fileclass_ReadString()
    Data.i @qDoc_fileclass_ReadUnicodeCharacter()
    Data.i @qDoc_fileclass_ReadWord()
    Data.i @qDoc_fileclass_WriteAsciiCharacter()
    Data.i @qDoc_fileclass_WriteByte()
    Data.i @qDoc_fileclass_WriteCharacter()
    Data.i @qDoc_fileclass_WriteData()
    Data.i @qDoc_fileclass_WriteDouble()
    Data.i @qDoc_fileclass_WriteFloat()
    Data.i @qDoc_fileclass_WriteInteger()
    Data.i @qDoc_fileclass_WriteLong()
    Data.i @qDoc_fileclass_WriteOLEString()
    Data.i @qDoc_fileclass_WriteQuad()
    Data.i @qDoc_fileclass_WriteString()
    Data.i @qDoc_fileclass_WriteUnicodeCharacter()
    Data.i @qDoc_fileclass_WriteWord()
    Data.i @qDoc_fileclass_ClearErrorFlag()
    Data.i @qDoc_fileclass_GetErrorFlag()
    Data.i @qDoc_fileclass_GetUserData()
    Data.i @qDoc_fileclass_SetUserData()
EndDataSection

CompilerEndIf

DisableExplicit

; IDE Options = PureBasic 5.21 LTS (Windows - x86)
; ExecutableFormat = Shared Dll
; CursorPosition = 9
; Folding = ---------
; Markers = 111
; EnableUnicode
; EnableXP
; Executable = nxReportU.dll